import logging, pytest, time
from common.Constants import DEFAULT_CONTEXT_UUID
from common.Settings import get_setting
from common.tools.object_factory.Connection import json_connection_id
from common.tools.object_factory.Device import json_device_id
from common.tools.object_factory.Service import json_service_id
from common.tools.grpc.Tools import grpc_message_to_json, grpc_message_to_json_string
from compute.tests.mock_osm.MockOSM import MockOSM
from context.client.ContextClient import ContextClient
from context.client.EventsCollector import EventsCollector
from context.proto.context_pb2 import (
    ConnectionEvent, ContextEvent, ContextId, DeviceEvent, Empty, EventTypeEnum, LinkEvent, ServiceEvent,
    TopologyEvent)
from .Objects import (
    CONTEXT_ID, CONTEXTS, DEVICE_O1_UUID, DEVICE_R1_UUID, DEVICE_R2_UUID, DEVICES, LINKS, TOPOLOGIES,
    WIM_MAPPING, WIM_PASSWORD, WIM_SERVICE_CONNECTION_POINTS, WIM_SERVICE_TYPE, WIM_USERNAME)

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)

EVENT_CREATE = EventTypeEnum.EVENTTYPE_CREATE
EVENT_UPDATE = EventTypeEnum.EVENTTYPE_UPDATE
EVENT_REMOVE = EventTypeEnum.EVENTTYPE_REMOVE


@pytest.fixture(scope='session')
def context_client():
    _client = ContextClient(get_setting('CONTEXTSERVICE_SERVICE_HOST'), get_setting('CONTEXTSERVICE_SERVICE_PORT_GRPC'))
    yield _client
    _client.close()


@pytest.fixture(scope='session')
def osm_wim():
    wim_url = 'http://{:s}:{:s}'.format(
        get_setting('COMPUTESERVICE_SERVICE_HOST'), str(get_setting('COMPUTESERVICE_SERVICE_PORT_HTTP')))
    return MockOSM(wim_url, WIM_MAPPING, WIM_USERNAME, WIM_PASSWORD)


def test_scenario_is_correct(context_client : ContextClient):  # pylint: disable=redefined-outer-name
    # ----- List entities - Ensure links are created -------------------------------------------------------------------
    response = context_client.ListContexts(Empty())
    assert len(response.contexts) == len(CONTEXTS)

    response = context_client.ListTopologies(ContextId(**CONTEXT_ID))
    assert len(response.topologies) == len(TOPOLOGIES)

    response = context_client.ListDevices(Empty())
    assert len(response.devices) == len(DEVICES)

    response = context_client.ListLinks(Empty())
    assert len(response.links) == len(LINKS)

    response = context_client.ListServices(ContextId(**CONTEXT_ID))
    assert len(response.services) == 0


def test_service_creation(context_client : ContextClient, osm_wim : MockOSM): # pylint: disable=redefined-outer-name
    # ----- Start the EventsCollector ----------------------------------------------------------------------------------
    events_collector = EventsCollector(context_client)
    events_collector.start()

    # ----- Create Service ---------------------------------------------------------------------------------------------
    service_uuid = osm_wim.create_connectivity_service(WIM_SERVICE_TYPE, WIM_SERVICE_CONNECTION_POINTS)
    osm_wim.get_connectivity_service_status(service_uuid)

    # ----- Validate collected events ----------------------------------------------------------------------------------
    EVENTS_EXPECTED = [
        (ServiceEvent,    EVENT_CREATE, json_service_id(service_uuid, context_id=CONTEXT_ID)),
        (ServiceEvent,    EVENT_UPDATE, json_service_id(service_uuid, context_id=CONTEXT_ID)),

        (DeviceEvent,     EVENT_UPDATE, json_device_id(DEVICE_O1_UUID)),
        (ServiceEvent,    EVENT_CREATE, json_service_id('{:s}:optical'.format(service_uuid), context_id=CONTEXT_ID)),
        (ConnectionEvent, EVENT_CREATE, json_connection_id('{:s}:optical:optical-line-system'.format(service_uuid))),

        (DeviceEvent,     EVENT_UPDATE, json_device_id(DEVICE_R1_UUID)),
        (DeviceEvent,     EVENT_UPDATE, json_device_id(DEVICE_R2_UUID)),
        (ServiceEvent,    EVENT_UPDATE, json_service_id(service_uuid, context_id=CONTEXT_ID)),
        (ConnectionEvent, EVENT_CREATE, json_connection_id('{:s}:emulated'.format(service_uuid))),
    ]
    events_received = events_collector.get_events(block=True, count=len(EVENTS_EXPECTED))
    for i,(event_expected, event_received) in enumerate(zip(EVENTS_EXPECTED, events_received)):
        LOGGER.info('event_expected[{:d}] = {:s}'.format(i, str(event_expected)))
        LOGGER.info('event_received[{:d}] = {:s}'.format(i, str(event_received)))
        assert isinstance(event_received, event_expected[0])
        entity_id_selector_function = {
            ContextEvent   : lambda event: event.context_id,
            TopologyEvent  : lambda event: event.topology_id,
            DeviceEvent    : lambda event: event.device_id,
            LinkEvent      : lambda event: event.link_id,
            ServiceEvent   : lambda event: event.service_id,
            ConnectionEvent: lambda event: event.connection_id,
        }.get(event_expected[0])
        entity_id = entity_id_selector_function(event_received)
        assert entity_id is not None
        assert event_received.event.event_type == event_expected[1]
        assert grpc_message_to_json(entity_id) == event_expected[2]

    # ----- Stop the EventsCollector -----------------------------------------------------------------------------------
    events_collector.stop()


def test_scenario_service_created(context_client : ContextClient):  # pylint: disable=redefined-outer-name
    # ----- List entities - Ensure service is created ------------------------------------------------------------------
    response = context_client.ListContexts(Empty())
    assert len(response.contexts) == len(CONTEXTS)

    response = context_client.ListTopologies(ContextId(**CONTEXT_ID))
    assert len(response.topologies) == len(TOPOLOGIES)

    response = context_client.ListDevices(Empty())
    assert len(response.devices) == len(DEVICES)

    response = context_client.ListLinks(Empty())
    assert len(response.links) == len(LINKS)

    response = context_client.ListServices(ContextId(**CONTEXT_ID))
    LOGGER.info('Services[{:d}] = {:s}'.format(len(response.services), grpc_message_to_json_string(response)))
    assert len(response.services) == 2 # L3NM + TAPI
    for service in response.services:
        service_id = service.service_id
        response = context_client.ListConnections(service_id)
        LOGGER.info('  ServiceId[{:s}] => Connections[{:d}] = {:s}'.format(
            grpc_message_to_json_string(service_id), len(response.connections), grpc_message_to_json_string(response)))
        assert len(response.connections) == 1 # one connection per service


def test_service_removal(context_client : ContextClient, osm_wim : MockOSM): # pylint: disable=redefined-outer-name
    # ----- Start the EventsCollector ----------------------------------------------------------------------------------
    events_collector = EventsCollector(context_client, log_events_received=True)
    events_collector.start()

    # ----- Delete Service ---------------------------------------------------------------------------------------------
    service_uuids = osm_wim.conn_info.keys()
    for service_uuid in service_uuids:
        osm_wim.delete_connectivity_service(service_uuid)

    # ----- Validate collected events ----------------------------------------------------------------------------------
    assert len(service_uuids) == 1  # assume a single service has been created
    service_uuid = set(service_uuids).pop()

    EVENTS_EXPECTED = [
        (ConnectionEvent, EVENT_REMOVE, json_connection_id('{:s}:emulated'.format(service_uuid))),
        (ServiceEvent,    EVENT_REMOVE, json_service_id(service_uuid, context_id=CONTEXT_ID)),
        (ConnectionEvent, EVENT_REMOVE, json_connection_id('{:s}:optical:optical-line-system'.format(service_uuid))),
        (ServiceEvent,    EVENT_REMOVE, json_service_id('{:s}:optical'.format(service_uuid), context_id=CONTEXT_ID)),
    ]
    events_received = events_collector.get_events(block=True, count=len(EVENTS_EXPECTED))
    for i,(event_expected, event_received) in enumerate(zip(EVENTS_EXPECTED, events_received)):
        LOGGER.info('event_expected[{:d}] = {:s}'.format(i, str(event_expected)))
        LOGGER.info('event_received[{:d}] = {:s}'.format(i, str(event_received)))
        assert isinstance(event_received, event_expected[0])
        entity_id_selector_function = {
            ContextEvent   : lambda event: event.context_id,
            TopologyEvent  : lambda event: event.topology_id,
            DeviceEvent    : lambda event: event.device_id,
            LinkEvent      : lambda event: event.link_id,
            ServiceEvent   : lambda event: event.service_id,
            ConnectionEvent: lambda event: event.connection_id,
        }.get(event_expected[0])
        entity_id = entity_id_selector_function(event_received)
        assert entity_id is not None
        assert event_received.event.event_type == event_expected[1]
        assert grpc_message_to_json(entity_id) == event_expected[2]

    # ----- Stop the EventsCollector -----------------------------------------------------------------------------------
    events_collector.stop()


def test_services_removed(context_client : ContextClient):  # pylint: disable=redefined-outer-name
    # ----- List entities - Ensure service is removed ------------------------------------------------------------------
    response = context_client.ListContexts(Empty())
    assert len(response.contexts) == len(CONTEXTS)

    response = context_client.ListTopologies(ContextId(**CONTEXT_ID))
    assert len(response.topologies) == len(TOPOLOGIES)

    response = context_client.ListDevices(Empty())
    assert len(response.devices) == len(DEVICES)

    response = context_client.ListLinks(Empty())
    assert len(response.links) == len(LINKS)

    response = context_client.ListServices(ContextId(**CONTEXT_ID))
    assert len(response.services) == 0
